unit ZREval;

{$B-}

interface

uses
  Windows, WinSpool, Sysutils, Messages, Classes, Controls, StdCtrls, ExtCtrls,
  ComCtrls, Buttons, Printers, Graphics, Forms, Dialogs, DB, DBTables;

type
  { TZRLibraryEntry }
  TZRLibraryItemClass = class of TObject;

  TZRLibraryEntry = class
    private
      fName        : String;
      fDescription : String;
      fData        : String;
      fVendor      : String;
      fItem        : TZRLibraryItemClass;
    public
      property Data        : String read fData write fData;
      property Description : String read fDescription write fDescription;
      property Name        : String read fName write fName;
      property Vendor      : String read fVendor write fVendor;
      property Item        : TZRLibraryItemClass read fItem write fItem;
  end;

  { TZRLibrary }
  TZRLibrary = class
    private
      fEntries : TStrings;
    protected
      function GetEntry(Index : integer) : TZRLibraryEntry; virtual;
    public
      constructor Create;
      destructor  Destroy; override;
      procedure Add(aItem : TZRLibraryItemClass; aName, aDescription, aVendor, aData: String);
      property Entries: TStrings read fEntries write fEntries;
      property Entry[Index: Integer] : TZRLibraryEntry read GetEntry; default;
  end;

  TZREvElement = class;

  { TZRFunctionLibrary }
  TZRFunctionLibrary = class(TZRLibrary)
    public
      function GetFunction(Name : String) : TZREvElement;
  end;

  { TZREvaluator related declarations }
  TZREvOperator = (opLess, opLessOrEqual, opGreater, opGreaterOrEqual, opEqual,
                   opUnequal, opPlus, opMinus, opOr, opMul, opDiv, opMod, opAnd);

  TZREvValueKind = (valInteger, valDouble, valString, valBoolean, valDateTime, valError);

  TZREvValue = record
    case Kind : TZREvValueKind of
      valInteger : (intResult : Longint);
      valDouble  : (dblResult : Double);
      valString  : (strResult : String[255]);
      valDateTime: (datResult : TDateTime);
      valBoolean : (booResult : Boolean);
  end;

  TZREvResult = class
    public
      Value : TZREvValue;
  end;

  TZRFiFo = class
    private
      fAggreg   : Boolean;
      fFiFo     : TList;
      fNextItem : Integer;
    public
      constructor Create;
      destructor  Destroy; override;
      procedure Start;
      procedure Put(Value : TObject);
      function  Get        : TObject;
      property Aggreg : Boolean read fAggreg write fAggreg;
  end;

  TZREvElement = class
    private
      fIsAggreg : Boolean;
    public
      constructor Create; virtual;
      function  Value(FiFo : TZRFiFo) : TZREvValue; virtual;
      procedure Reset; virtual;
      property IsAggreg : Boolean read fIsAggreg write fIsAggreg;
  end;

  TZREvElementFunction = class(TZREvElement)
    private
    protected
      ArgList : TList;
      function  ArgumentOK(Value : TZREvElement) : Boolean;
      function  Argument(Index : Integer) : TZREvValue;
      procedure FreeArguments;
      procedure GetArguments(FiFo : TZRFiFo);
      procedure Aggregate; virtual;
      function  Calculate : TZREvValue; virtual;
    public
      constructor Create; override;
      destructor Destroy; override;
      function Value(FiFo : TZRFiFo) : TZREvValue; override;
  end;

  TZREvElementFunctionClass = class of TZREvElementFunction;

  TZREvElementArgumentEnd = class(TZREvElement);

  TZREvElementDataField = class(TZREvElement)
    private
      fDataSet : TDataSet;
      fFieldNo : Integer;
      fField   : TField;
    public
      constructor CreateField(aField : TField); virtual;
      function Value(FiFo : TZRFiFo) : TZREvValue; override;
  end;

  TZREvElementError = class(TZREvElement)
    public
      function Value(FiFo : TZRFiFo) : TZREvValue; override;
  end;

  { TZREvaluator class }

  TZREvaluator = class (TObject)
    private
      fDataSets  : TList;
      FiFo       : TZRFiFo;
      fPrepared  : Boolean;
      function  EvalFunctionExpr(const strFunc : String) : TZREvValue;
      function  EvalSimpleExpr(const strSimplExpr : String) : TZREvValue;
      function  EvalTerm(const strTermExpr : String) : TZREvValue;
      function  EvalFactor(strFactorExpr : String) : TZREvValue;
      function  EvalString(const strString : String) : TZREvValue;
      function  EvalConstant(const strConstant : String) : TZREvValue;
      function  GetAggregate : Boolean;
      function  NegateResult(const Res : TZREvValue) : TZREvValue;
      function  Evaluate(const strExpr : String) : TZREvValue;
      procedure FindDelimiter(strArg : String; var Pos : Integer);
      procedure FindOp1(const strExpr : String; var Op : TZREvOperator; var Pos, Len : Integer);
      procedure FindOp2(const strExpr : String; var Op : TZREvOperator; var Pos, Len : Integer);
      procedure FindOp3(const strExpr : String; var Op : TZREvOperator; var Pos, Len : Integer);
      procedure SetAggregate(Value : Boolean);
      procedure TrimString(var strString : String);
    protected
      function EvalFunction(strFunc : String; const strArg : String) : TZREvValue; virtual;
      function EvalVariable(strVariable : String) : TZREvValue; virtual;
      function GetIsAggreg : Boolean;
    public
      constructor Create;
      destructor Destroy; override;
      function  Calculate(const StrExpr : String) : TZREvValue;
      function  Value : TZREvValue;
      procedure Prepare(const StrExpr : String);
      procedure Reset;
      procedure UnPrepare;
      property IsAggreg  : Boolean read GetIsAggreg;
      property Aggregate : Boolean read GetAggregate write SetAggregate;
      property DataSets  : TList   read fDataSets write fDataSets;
      property Prepared  : Boolean read fPrepared write fPrepared;
  end;

function Calculate(Formula: String; Values: array of const): Variant;

procedure RegisterZRFunction(FunctionClass : TZRLibraryItemClass; Name, Description, Vendor, Arguments : String);

var
  ZRFunctionLibrary : TZRFunctionLibrary;

implementation

uses
  ZRConst;

function Calculate(Formula: String; Values: array of const): Variant;
var
  E: TZREvaluator;
  R: TZREvValue;
begin
  Formula:= Format(Formula, Values);
  E:= TZREvaluator.Create;
  try
    R:= E.Calculate(Formula);
    case R.Kind of
      valInteger : Result:= R.intResult;
      valDouble  : Result:= R.dblResult;
      valString  : Result:= R.strResult;
      valBoolean : Result:= R.booResult;
      valDateTime: Result:= R.datResult;
      else        Result:= Null;
    end;
  except
    E.Free;
    VarClear(Result);
  end;
end;


const
  ArgSeparator : Char = ',';

procedure RegisterZRFunction(FunctionClass : TZRLibraryItemClass; Name, Description, Vendor, Arguments : string);
begin
  ZRFunctionLibrary.Add(FunctionClass, Name, Description, Vendor, Arguments);
end;

{ TZRLibrary }

constructor TZRLibrary.Create;
begin
  inherited Create;
  fEntries := TStringList.Create;
end;

destructor TZRLibrary.Destroy;
var
  I : integer;
begin
  for I := 0 to Entries.Count - 1 do
    fEntries.Objects[I].Free;
  fEntries.Free;
  inherited Destroy;
end;

procedure TZRLibrary.Add(aItem : TZRLibraryItemClass; aName, aDescription, aVendor, aData: String);
var
  aEntry : TZRLibraryEntry;
begin
  aEntry := TZRLibraryEntry.Create;
  with aEntry do
    begin
      Name        := aName;
      Description := aDescription;
      Vendor      := aVendor;
      Item        := aItem;
      Data        := aData;
    end;
  fEntries.AddObject(aName, aEntry);
end;

function TZRLibrary.GetEntry(Index : integer) : TZRLibraryEntry;
begin
  if Index <= fEntries.Count then
    Result := TZRLibraryEntry(fEntries.Objects[Index])
  else
    Result := nil;
end;

{ TZRFunctionLibrary }

function TZRFunctionLibrary.GetFunction(Name : String) : TZREvElement;
var
  I             : integer;
  AObject       : TZREvElementFunctionClass;
  aLibraryEntry : TZRLibraryEntry;
begin
  I := Entries.IndexOf(Name);
  if I>=0 then
    begin
      aLibraryEntry := TZRLibraryEntry(Entry[I]);
      aObject := TZREvElementFunctionClass(aLibraryEntry.Item);
      Result := aObject.Create;
    end
  else
    Result := TZREvElementError.Create;
end;

{ TZREvaluator }

constructor TZRFiFo.Create;
begin
  fFiFo     := TList.Create;
  fAggreg   := False;
  fNextItem := 0;
end;

destructor TZRFiFo.Destroy;
var
  I : integer;
begin
  for I := 0 to fFiFo.Count-1 do
    TObject(fFiFo[I]).Free;
  fFiFo.Free;
  inherited Destroy;
end;

procedure TZRFiFo.Start;
begin
  fNextItem := 0;
end;

procedure TZRFiFo.Put(Value : TObject);
begin
  fFiFo.Add(Value);
end;

function TZRFiFo.Get : TObject;
begin
  if fNextItem < fFiFo.Count then
  begin
    Result := fFiFo[fNextItem];
    Inc(fNextItem);
  end else
    Result := nil;
end;

{ TZREvElement }

constructor TZREvElement.Create;
begin
  inherited Create;
  fIsAggreg := False;
end;

function TZREvElement.Value(FiFo : TZRFiFo) : TZREvValue;
begin
end;

procedure TZREvElement.Reset;
begin
end;

{ TZREvElementOperator }

type
  TZREvElementOperator = class(TZREvElement)
  private
    fCode :  TZREvOperator;
    procedure ConvertZREvResults(var Res1 : TZREvValue; var Res2 : TZREvValue);
  public
    constructor CreateOperator(Code : TZREvOperator);
    function Value(FiFo : TZRFiFo) : TZREvValue; override;
  end;

constructor TZREvElementOperator.CreateOperator(Code : TZREvOperator);
begin
  inherited Create;
  fCode := Code;
end;

procedure TZREvElementOperator.ConvertZREvResults(var Res1 : TZREvValue; var Res2 : TZREvValue);
begin
  if Res1.Kind <> Res2.Kind then
    begin
      if ((Res1.Kind = valInteger) and (Res2.Kind = valDouble)) then
        begin
          Res1.Kind      := valDouble;
          Res1.dblResult := Res1.intResult;
        end else
      if ((Res1.Kind = valDouble) and (Res2.Kind = valInteger)) then
        begin
          Res2.Kind      := valDouble;
          Res2.dblResult := Res2.intResult;
        end
      else
        begin
          Res1.Kind := valError;
          Res2.Kind := valError;
        end;
    end;
end;

function TZREvElementOperator.Value(FiFo : TZRFiFo) : TZREvValue;
var
  Res1,
  Res2 : TZREvValue;
begin
  Res1 := TZREvElement(FiFo.Get).Value(FiFo);
  Res2 := TZREvElement(FiFo.Get).Value(FiFo);
  ConverTZREvResults(Res1, Res2);
  Result.Kind := Res1.Kind;
  if Result.Kind <> valError then
    case fCode of
      opPlus: case Result.Kind of
                valInteger : Result.intResult := Res1.intResult + Res2.intResult;
                valDouble  : Result.dblResult := Res1.dblResult + Res2.dblResult;
                valDateTime: Result.datResult := Res1.datResult + Res2.datResult;
                valString  : Result.strResult := Res1.strResult + Res2.strResult;
                valBoolean : Result.Kind := valError;
              end;
      opMinus:  case Result.Kind of
                  valInteger : Result.intResult := Res1.intResult - Res2.intResult;
                  valDouble  : Result.dblResult := Res1.dblResult - Res2.dblResult;
                  valDateTime: Result.datResult := Res1.datResult - Res2.datResult;
                  valString  : Result.Kind := valError;
                  valBoolean : Result.Kind := valError;
                end;
      opMul: case Result.Kind of
               valInteger : Result.intResult := Res1.intResult * Res2.intResult;
               valDouble  : Result.dblResult := Res1.dblResult * Res2.dblResult;
               valString  : Result.Kind := valError;
               valDateTime: Result.Kind := valError;
               valBoolean : Result.Kind := valError;
             end;
      opMod: case Result.Kind of
               valInteger : if Res2.intResult <> 0 then
                              Result.intResult := Res1.intResult mod Res2.intResult
                            else
                              Result.Kind := valError;
               valDouble  : Result.Kind := valError;
               valString  : Result.Kind := valError;
               valDateTime: Result.Kind := valError;
               valBoolean : Result.Kind := valError;
             end;
      opDiv: case Result.Kind of
               valInteger : if Res2.intResult <> 0 then
                              Result.intResult := Res1.intResult div Res2.intResult
                            else
                              Result.Kind := valError;
               valDouble  : if Res2.dblResult <> 0 then
                              Result.dblResult := Res1.dblResult / Res2.dblResult
                            else
                              Result.Kind := valError;
               valDateTime: Result.Kind := valError;
               valString  : Result.Kind := valError;
               valBoolean : Result.Kind := valError;
             end;
      opGreater: begin
                   Result.Kind := valBoolean;
                   case Res1.Kind of
                     valInteger : Result.booResult := Res1.intResult > Res2.intResult;
                     valDouble  : Result.booResult := Res1.dblResult > Res2.dblResult;
                     valString  : Result.booResult := Res1.strResult > Res2.strResult;
                     valDateTime: Result.booResult := Res1.datResult > Res2.datResult;
                     valBoolean : Result.Kind := valError;
                   end;
                 end;
      opGreaterOrEqual: begin
                Result.Kind := valBoolean;
                case Res1.Kind of
                  valInteger : Result.booResult := Res1.intResult >= Res2.intResult;
                  valDouble  : Result.booResult := Res1.dblResult >= Res2.dblResult;
                  valString  : Result.booResult := Res1.strResult >= Res2.strResult;
                  valDateTime: Result.booResult := Res1.datResult >= Res2.datResult;
                  valBoolean : Result.Kind := valError;
                end;
             end;
      opLess: begin
                Result.Kind := valBoolean;
                case Res1.Kind of
                  valInteger : Result.booResult := Res1.intResult < Res2.intResult;
                  valDouble  : Result.booResult := Res1.dblResult < Res2.dblResult;
                  valString  : Result.booResult := Res1.strResult < Res2.strResult;
                  valDateTime: Result.booResult := Res1.datResult < Res2.datResult;
                  valBoolean : Result.Kind := valError;
                end;
              end;
      opLessOrEqual: begin
                 Result.Kind := valBoolean;
                 case Res1.Kind of
                   valInteger : Result.booResult := Res1.intResult <= Res2.intResult;
                   valDouble  : Result.booResult := Res1.dblResult <= Res2.dblResult;
                   valString  : Result.booResult := Res1.strResult <= Res2.strResult;
                   valDateTime: Result.booResult := Res1.datResult <= Res2.datResult;
                   valBoolean : Result.Kind := valError;
                 end;
               end;
      opEqual: begin
                 Result.Kind := valBoolean;
                 case Res1.Kind of
                   valInteger : Result.booResult := Res1.intResult = Res2.intResult;
                   valDouble  : Result.booResult := Res1.dblResult = Res2.dblResult;
                   valString  : Result.booResult := Res1.strResult = Res2.strResult;
                   valDateTime: Result.booResult := Res1.datResult = Res2.datResult;
                   valBoolean : Result.booResult := Res1.booResult = Res2.booResult;
                 end;
               end;
      opUnequal: begin
                   Result.Kind := valBoolean;
                   case Res1.Kind of
                     valInteger : Result.booResult := Res1.intResult <> Res2.intResult;
                     valDouble  : Result.booResult := Res1.dblResult <> Res2.dblResult;
                     valString  : Result.booResult := Res1.strResult <> Res2.strResult;
                     valBoolean : Result.booResult := Res1.booResult <> Res2.booResult;
                     valDateTime: Result.booResult := Res1.datResult <> Res2.datResult;
                   end;
                 end;
      opOr: begin
              Result.Kind := Res1.Kind;
              case Res1.Kind of
                valInteger : Result.intResult := Res1.intResult or Res2.intResult;
                valDouble  : Result.Kind := valError;
                valDateTime: Result.Kind := valError;
                valString  : Result.Kind := valError;
                valBoolean : Result.booResult := Res1.booResult or Res2.booResult;
              end;
            end;
      opAnd: begin
               Result.Kind := Res1.Kind;
               case Res1.Kind of
                 valInteger : Result.intResult := Res1.intResult and Res2.intResult;
                 valDouble  : Result.Kind := valError;
                 valDateTime: Result.Kind := valError;
                 valString  : Result.Kind := valError;
                 valBoolean : Result.booResult := Res1.booResult and Res2.booResult;
               end;
             end;
    end;
end;

{ TZREvElementConstant }

type
  TZREvElementConstant = class(TZREvElement)
  private
    fValue : TZREvValue;
  public
    constructor CreateConstant(Value : TZREvValue);
    function Value(FiFo : TZRFiFo) : TZREvValue; override;
  end;

constructor TZREvElementConstant.CreateConstant(Value : TZREvValue);
begin
  inherited Create;
  fValue := Value;
end;

function TZREvElementConstant.Value(FiFo : TZRFiFo): TZREvValue;
begin
  Result := fValue;
end;

{ TZREvElementString }

type
  TZREvElementString = class(TZREvElement)
  private
    fValue : String;
  public
    constructor CreateString(Value : String);
    function Value(FiFo : TZRFiFo) : TZREvValue; override;
  end;

constructor TZREvElementString.CreateString(Value : String);
begin
  inherited Create;
  fValue := Value;
end;

function TZREvElementString.Value(FiFo : TZRFiFo) : TZREvValue;
begin
  Result.Kind      := valString;
  Result.strResult := fValue;
end;

{ TZREvElementFunction }

constructor TZREvElementFunction.Create;
begin
  inherited Create;
  ArgList := TList.Create;
end;

destructor TZREvElementFunction.Destroy;
begin
  ArgList.Free;
  inherited Destroy;
end;

procedure TZREvElementFunction.GetArguments(FiFo : TZRFiFo);
var
  aArgument: TZREvElement;
  AResult  : TZREvResult;
begin
  repeat
    aArgument := TZREvElement(FiFo.Get);
    if not (aArgument is TZREvElementArgumentEnd) then
    begin
      aResult := TZREvResult.Create;
      aResult.Value := aArgument.Value(FiFo);
      ArgList.Add(aResult);
    end;
  until aArgument is TZREvElementArgumentEnd;
end;

procedure TZREvElementFunction.FreeArguments;
var
  I : Integer;
begin
  for I := 0 to ArgList.Count - 1 do
    TZREvElement(ArgList.Items[I]).Free;
  ArgList.Clear;
end;

function TZREvElementFunction.Argument(Index : integer): TZREvValue;
begin
  if Index <= ArgList.Count then
    Result := TZREvResult(ArgList[Index]).Value;
end;

function TZREvElementFunction.Value(FiFo : TZRFiFo) : TZREvValue;
begin
  GetArguments(FiFo);
  if FiFo.Aggreg then
    Aggregate;
  Result := Calculate;
  FreeArguments;
end;

function TZREvElementFunction.ArgumentOK(Value : TZREvElement) : Boolean;
begin
  Result := not (Value is TZREvElementArgumentEnd) and not (Value is TZREvElementError);
end;

procedure TZREvElementFunction.Aggregate;
begin
end;

function TZREvElementFunction.Calculate : TZREvValue;
begin
  Result.Kind := valError;
end;

{ TZREvTrue }

type
  TZREvTrue = class(TZREvElementFunction)
  public
    function Calculate : TZREvValue; override;
  end;

  TZREvFalse = class(TZREvElementFunction)
  public
    function Calculate : TZREvValue; override;
  end;

function TZREvTrue.Calculate : TZREvValue;
begin
  if ArgList.Count = 0 then
  begin
    Result.Kind := valBoolean;
    Result.booResult := True;
  end else
    Result.Kind := valError;
end;

function TZREvFalse.Calculate : TZREvValue;
begin
  if ArgList.Count = 0 then
  begin
    Result.Kind := valBoolean;
    Result.booResult := False;
  end else
    Result.Kind := valError;
end;

{ TZREvIfFunction }

type
  TZREvIfFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvValue; override;
  end;

function TZREvIfFunction.Calculate : TZREvValue;
begin
  if (ArgList.Count = 3) and (Argument(0).Kind = valBoolean) then
  begin
    if Argument(0).booResult then
      Result := Argument(1)
    else
      Result := Argument(2);
  end else
    result.Kind := valError;
end;

{ TZREvTypeOfFunction }

type
  TZREvTypeOfFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvValue; override;
  end;

function TZREvTypeOfFunction.Calculate : TZREvValue;
begin
  Result.Kind := valString;
  if ArgList.Count = 1 then
  begin
    case Argument(0).Kind of
      valInteger : Result.strResult := 'INTEGER';   {<-- do not resource }
      valDouble  : Result.strResult := 'FLOAT';     {<-- do not resource }
      valString  : Result.strResult := 'STRING';    {<-- do not resource }
      valBoolean : Result.strResult := 'BOOLEAN';   {<-- do not resource }
      valDateTime: Result.strResult := 'DATE';      {<-- do not resource }
      valError   : Result.strResult := 'ERROR';     {<-- do not resource }
    else
      Result.Kind := valError;
    end
  end else
    Result.Kind := valError;
end;

{ TZREvIntFunction }

type
  TZREvIntFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvValue; override;
  end;

function TZREvIntFunction.Calculate : TZREvValue;
begin
  Result.Kind := valInteger;
  if ArgList.Count = 1 then
  begin
    case Argument(0).Kind of
      valInteger: Result.intResult := Argument(0).intResult;
      valDouble : Result.intResult := Round(Int(Argument(0).dblResult));
    else
      Result.Kind := valError;
    end
  end else
    Result.Kind := valError;
end;

{ TZREvFracFunction }

type
  TZREvFracFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvValue; override;
  end;

function TZREvFracFunction.Calculate : TZREvValue;
begin
  Result.Kind := valDouble;
  if ArgList.Count = 1 then
  begin
    case Argument(0).Kind of
      valInteger: Result.dblResult := 0;
      valDouble : Result.dblResult := Frac(Argument(0).dblResult);
    else
      Result.Kind := valError;
    end
  end else
    Result.Kind := valError;
end;

{ TZREvSQRTFunction }

type
  TZREvSQRTFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvValue; override;
  end;

function TZREvSQRTFunction.Calculate : TZREvValue;
begin
  Result.Kind := valDouble;
  if ArgList.Count = 1 then
  begin
    case Argument(0).Kind of
      valInteger: Result.dblResult := SQRT(Argument(0).intResult);
      valDouble : Result.dblResult := SQRT(Argument(0).dblResult);
    else
      Result.Kind := valError;
    end
  end else
    Result.Kind := valError;
end;

{ TZREvStrFunction }

type
  TZREvStrFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvValue; override;
  end;

function TZREvStrFunction.Calculate : TZREvValue;
var
  ArgRes : TZREvValue;
begin
  if ArgList.Count = 1 then
  begin
    ArgRes := Argument(0);
    Result.Kind := valString;
    case ArgRes.Kind of
      valInteger : Result.strResult := IntToStr(ArgRes.intResult);
      valDouble  : Result.strResult := FloatToStr(ArgRes.dblResult);
      valBoolean : if ArgRes.booResult then
                     Result.strResult := 'True'          {<-- do not resource }
                   else
                     Result.strResult := 'False';        {<-- do not resource }
      valDateTime: Result.strResult := DateTimeToStr(ArgRes.datResult);
    else
      Result.Kind := valError;
    end
  end else
    Result.Kind := valError;
end;

{ TZREvUpperFunction }

type
  TZREvUpperFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvValue; override;
  end;

function TZREvUpperFunction.Calculate : TZREvValue;
begin
  if (ArgList.Count = 1) and (Argument(0).Kind = valString) then
  begin
    Result.Kind := valString;
    Result.strResult := AnsiUpperCase(Argument(0).strResult);
  end else
    Result.Kind := valError;
end;

{ TZREvLowerFunction }

type
  TZREvLowerFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvValue; override;
  end;

function TZREvLowerFunction.Calculate : TZREvValue;
begin
  if (ArgList.Count = 1) and (Argument(0).Kind = valString) then
  begin
    Result.Kind := valString;
    Result.strResult := AnsiLowerCase(Argument(0).strResult);
  end else
    Result.Kind := valError;
end;

{ TZREvPrettyFunction }

type
  TZREvPrettyFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvValue; override;
  end;

function TZREvPrettyFunction.Calculate : TZREvValue;
begin
  if (ArgList.Count = 1) and (Argument(0).Kind = valString) then
  begin
    Result.Kind := valString;
    Result.strResult := AnsiUpperCase(Copy(Argument(0).strResult, 1, 1)) +
                        AnsiLowerCase(Copy(Argument(0).strResult, 2, Length(Argument(0).strResult)));
  end else
    Result.Kind := valError;
end;

{ TZREvTimeFunction }

type
  TZREvTimeFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvValue; override;
  end;

function TZREvTimeFunction.Calculate : TZREvValue;
begin
  if ArgList.Count = 0 then
  begin
    Result.Kind      := valString;
    Result.strResult := TimeToStr(Now);
  end else
    Result.Kind := valError;
end;

{ TZREvTimeFunction }

type
  TZREvDateFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvValue; override;
  end;

function TZREvDateFunction.Calculate : TZREvValue;
begin
  if ArgList.Count=0 then
  begin
    Result.Kind      := valString;
    Result.StrResult := DateToStr(Date);
  end else
    Result.Kind := valError;
end;

{ TZREvCopyFunction }

type
  TZREvCopyFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvValue; override;
  end;

function TZREvCopyFunction.Calculate : TZREvValue;
begin
  if (ArgList.Count = 3) and
    (Argument(0).Kind = valString) and
    (Argument(1).Kind = valInteger) and
    (Argument(2).Kind = valInteger) then
  begin
    Result.strResult := copy(Argument(0).strResult, Argument(1).intResult, Argument(2).intResult);
    Result.Kind := valString;
  end else
    Result.Kind := valError;
end;

{ TZREvFormatFloatFunction }

type
  TZREvFormatFloatFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvValue; override;
  end;

function TZREvFormatFloatFunction.Calculate : TZREvValue;
begin
  if (ArgList.Count = 2) and
     ((Argument(1).Kind = valInteger) or (Argument(1).Kind = valDouble)) and
     (Argument(0).Kind = valString) then
  begin
    Result.Kind := valString;
    if Argument(1).Kind = valInteger then
      Result.StrResult := FormatFloat(Argument(0).strResult, Argument(1).intResult * 1.0)
    else
      Result.StrResult := FormatFloat(Argument(0).strResult, Argument(1).dblResult);
  end else
    Result.Kind := valError;
end;

{ TZREvSumFunction }

type
  TZREvSumFunction = class(TZREvElementFunction)
  private
    SumResult   : TZREvValue;
    ResAssigned : Boolean;
  public
    constructor Create; override;
    procedure  Aggregate; override;
    function   Calculate : TZREvValue; override;
    procedure  Reset; override;
  end;

constructor TZREvSumFunction.Create;
begin
  inherited Create;
  ResAssigned := False;
  IsAggreg    := True;
end;

procedure TZREvSumFunction.Reset;
begin
  ResAssigned := False;
end;

procedure TZREvSumFunction.Aggregate;
var
  aValue : TZREvValue;
begin
  if ArgList.Count = 1 then
  begin
    aValue := Argument(0);
    if ResAssigned then
      case aValue.Kind of
        valInteger: SumResult.IntResult := SumResult.intResult + aValue.intResult;
        valDouble : SumResult.dblResult := SumResult.dblResult + aValue.dblResult;
        else        SumResult.Kind := valError;
      end else
    begin
      SumResult.Kind := aValue.Kind;
      case aValue.Kind of
        valInteger: SumResult.intResult := aValue.intResult;
        valDouble : SumResult.dblResult := aValue.dblResult;
        else        SumResult.Kind := valError;
      end;
    end;
    ResAssigned := True;
  end else
    SumResult.Kind := valError;
end;

function TZREvSumFunction.Calculate : TZREvValue;
begin
  if ResAssigned then
    Result := SumResult
  else
    Result.Kind := valError;
end;

{ TZREvAverageFunction }

type
  TZREvAverageFunction = class(TZREvSumFunction)
  private
    Count   : Longint;
    aResult : TZREvValue;
  public
    function  Calculate : TZREvValue; override;
    procedure Aggregate; override;
    procedure Reset; override;
  end;

procedure TZREvAverageFunction.Reset;
begin
  inherited Reset;
  aResult.Kind := valError;
  Count        := 0;
  IsAggreg     := True;
end;

procedure TZREvAverageFunction.Aggregate;
var
  aValue : TZREvValue;
begin
  inherited Aggregate;
  Inc(Count);
  aValue       := inherited Calculate;
  aResult.Kind := valDouble;
  case aValue.Kind of
    valInteger: aResult.dblResult := aValue.intResult / Count;
    valDouble : aResult.dblResult := aValue.dblResult / Count;
  else
    aResult.Kind := valError;
  end;
end;

function TZREvAverageFunction.Calculate : TZREvValue;
begin
  Result := aResult
end;

{ TZREvMaxFunction }

type
  TZREvMaxFunction = class(TZREvElementFunction)
  private
    MaxResult   : TZREvValue;
    ResAssigned : Boolean;
  public
    constructor Create; override;
    function  Calculate : TZREvValue; override;
    procedure Aggregate; override;
    procedure Reset; override;
  end;

constructor TZREvMaxFunction.Create;
begin
  inherited Create;
  ResAssigned := False;
  IsAggreg    := True;
end;

procedure TZREvMaxFunction.Reset;
begin
  ResAssigned := False;
end;

procedure TZREvMaxFunction.Aggregate;
var
  aValue : TZREvValue;
begin
  if ArgList.Count = 1 then
  begin
    aValue := Argument(0);
    if ResAssigned then
    begin
      case MaxResult.Kind of
        valInteger : if aValue.intResult > MaxResult.intResult then
                       MaxResult.intResult := aValue.intResult;
        valDouble  : if aValue.dblResult > MaxResult.dblResult then
                       MaxResult.dblResult := aValue.dblResult;
        valString  : if aValue.strResult > MaxResult.strResult then
                       MaxResult.strResult := aValue.strResult;
        valDateTime: if aValue.datResult > MaxResult.datResult then
                       MaxResult.datResult := aValue.datResult;
        valBoolean : if aValue.booResult > MaxResult.booResult then
                       MaxResult.booResult := aValue.booResult;
      else
        MaxResult.Kind := valError;
      end
    end else
    begin
      MaxResult   := aValue;
      ResAssigned := True;
    end
  end else
    MaxResult.Kind := valError;
end;

function TZREvMaxFunction.Calculate : TZREvValue;
begin
  if ResAssigned then
    Result := MaxResult
  else
    Result.Kind := valError;
end;

{ TZREvMinFunction }

type
  TZREvMinFunction = class(TZREvElementFunction)
  private
    MinResult   : TZREvValue;
    ResAssigned : Boolean;
  public
    constructor Create; override;
    function  Calculate : TZREvValue; override;
    procedure Aggregate; override;
    procedure Reset; override;
  end;

constructor TZREvMinFunction.Create;
begin
  inherited Create;
  ResAssigned := False;
  IsAggreg    := True;
end;

procedure TZREvMinFunction.Reset;
begin
  ResAssigned := False;
end;

procedure TZREvMinFunction.Aggregate;
var
  aValue : TZREvValue;
begin
  if ArgList.Count = 1 then
  begin
    aValue := Argument(0);
    if ResAssigned then
    begin
      case MinResult.Kind of
        valInteger : if aValue.intResult < MinResult.intResult then
                       MinResult.intResult := aValue.intResult;
        valDouble  : if aValue.dblResult < MinResult.dblResult then
                       MinResult.dblResult := aValue.dblResult;
        valString  : if aValue.strResult < MinResult.strResult then
                       MinResult.strResult := aValue.strResult;
        valDateTime: if aValue.datResult < MinResult.datResult then
                       MinResult.datResult := aValue.datResult;
        valBoolean : if aValue.booResult > MinResult.booResult then
                       MinResult.booResult := aValue.booResult;
      else
        MinResult.Kind := valError;
      end
    end else
    begin
      MinResult   := aValue;
      ResAssigned := True;
    end
  end else
    MinResult.Kind := valError;
end;

function TZREvMinFunction.Calculate : TZREvValue;
begin
  if ResAssigned then
    Result := MinResult
  else
    Result.Kind := valError;
end;

{ TZREvCountFunction }

type
  TZREvCountFunction = class(TZREvElementFunction)
  private
    fCount : Integer;
  public
    constructor Create; override;
    function  Value(FiFo : TZRFiFo) : TZREvValue; override;
    procedure Reset; override;
  end;

constructor TZREvCountFunction.Create;
begin
  inherited Create;
  fCount   := 0;
  IsAggreg := True;
end;

function TZREvCountFunction.Value(FiFo : TZRFiFo) : TZREvValue;
begin
  GetArguments(FiFo);
  if (ArgList.Count = 0) then
  begin
    if FiFo.Aggreg then
      Inc(fCount);
    Result.Kind      := valInteger;
    Result.intResult := fCount;
  end else
    Result.Kind := valError;
  FreeArguments;
end;

procedure TZREvCountFunction.Reset;
begin
  fCount := 0;
end;

{ TZREvElementDataField }

constructor TZREvElementDataField.CreateField(aField : TField);
begin
  inherited Create;
  fDataSet := aField.DataSet;
  fFieldNo := aField.Index;
  fField := aField;
end;

function TZREvElementDataField.Value(FiFo : TZRFiFo) : TZREvValue;
begin
  if fDataSet.DefaultFields then
    fField := fDataSet.Fields[fFieldNo];
  if fField is TStringField then
  begin
    Result.Kind := valString;
    Result.strResult := TStringField(fField).Value;
  end else
    if (fField is TIntegerField) or
       (fField is TSmallIntField) or
       (fField is TWordField) then
    begin
      Result.Kind := valInteger;
      Result.intResult := fField.AsInteger;
    end else
      if (fField is TFloatField)  or
         (fField is TCurrencyField) or
         (fField is TBCDField) then
      begin
        Result.Kind := valDouble;
        Result.dblResult := TFloatField(fField).AsFloat;
      end else
        if fField is TBooleanField then
        begin
          Result.Kind := valBoolean;
          Result.BooResult := TBooleanField(fField).Value;
        end else
          if (fField is TDateField) or
             (fField is TTimeField) or
             (fField is TDateTimeField) then
          begin
            Result.Kind := valDateTime;
            Result.datResult := fField.AsDateTime;
          end else
            Result.Kind := valError
end;

function TZREvElementError.Value(FiFo : TZRFiFo) : TZREvValue;
begin
  Result.Kind := valError;
end;

{ TZREvaluator }

constructor TZREvaluator.Create;
begin
  Prepared := False;
end;

destructor TZREvaluator.Destroy;
begin
  if Prepared then Unprepare;
  inherited Destroy;
end;

procedure TZREvaluator.TrimString(var strString : string);
var
  intStart,
  intEnd : integer;
begin
  intStart := 1;
  intEnd := Length(strString);
  while (copy(strString, intStart,1) = ' ') and (intStart < intEnd) do
    inc(intStart);
  while (copy(strString, intEnd,1) = ' ') and (intEnd > intStart) do
    dec(intEnd);
  strString := Copy(strString, intStart, intEnd - intStart + 1);
end;

procedure TZREvaluator.FindDelimiter(strArg : string; var Pos : integer);
var
  n             : integer;
  FoundDelim    : boolean;
  booString     : boolean;
  intParenteses : integer;
begin
  if strArg='' then
    Pos := 0
  else
  begin
    FoundDelim := false;
    BooString := false;
    intParenteses := 0;
    N := 1;
    while (N<length(strArg)) and not FoundDelim do
    begin
      case StrArg[N] of
        '(' : if not booString then inc(intParenteses);
        ')' : if not booString then dec(intParenteses);
        '''' : booString := not booString;
      end;
      if (intParenteses=0) and not booString then
        if strArg[N]=ArgSeparator then
        begin
          FoundDelim := true;
          break;
        end;
      inc(N);
    end;
    if FoundDelim then
      Pos := N
    else
      Pos := 0;
  end;
end;

procedure TZREvaluator.FindOp1(const strExpr : string; var Op : TZREvOperator; var Pos, Len : integer);
var
  n : integer;
  booFound : boolean;
  intParenteses : integer;
  aString : string[255];
  booString : boolean;
begin
  n := 1;
  intParenteses := 0;
  booFound := false;
  Len := 1;
  aString := strExpr;
  booString := false;
  while (n < Length(strExpr)) and (not booFound) do
  begin
    booFound := true;
    case aString[N] of
      '(' : if not booString then inc(intParenteses);
      ')' : if not booString then dec(intParenteses);
      '''' : booString := not booString;
    end;
    if (intParenteses = 0) and (n > 1) and not booString then
      case aString[N] of
        '<' : begin
                if aString[N + 1] = '>' then
                begin
                  Op := opUnequal;
                  Len := 2;
                end else
                  if aString[N + 1] = '=' then
                  begin
                    Op := opLessOrEqual;
                    Len := 2;
                  end else
                    Op := opLess;
                end;
        '>' : if aString[N + 1] = '=' then
              begin
                Op := opGreaterOrEqual;
                Len := 2;
              end else
                Op := opGreater;
        '=' : Op := opEqual;
      else
        booFound := false;
      end
    else
      booFound := false;
    inc(N);
  end;
  if booFound then
    Pos := n - 1
  else
    Pos := -1;
end;

procedure TZREvaluator.FindOp2(const strExpr : string; var Op : TZREvOperator; var Pos, Len : integer);
var
  n : integer;
  booFound : boolean;
  intParenteses : integer;
  booString : boolean;
  aString : string[255];
begin
  n := 1;
  intParenteses := 0;
  booFound := false;
  booString := false;
  aString := strExpr;
  Len := 1;
  while (n < Length(strExpr)) and (not booFound) do
  begin
    booFound := true;
    case aString[N] of
      '(' : if not boostring then inc(intParenteses);
      ')' : if not boostring then dec(intParenteses);
      '''': booString := not booString;
    end;
    if (intParenteses = 0) and (not booString) and (N > 1) then
      case aString[N] of
        '+' : Op := opPlus;
        '-' : Op := opMinus;
        ' ' : if (AnsiLowercase(copy(strExpr, N + 1, 3)) = 'or ') then
              begin
                Op := opOr;
                Len := 2;
                inc(N);
              end else
                booFound := false;
      else
        booFound := false;
    end else
      booFound := false;
    inc(N);
  end;
  if booFound then
    Pos := N - 1
  else
    Pos := -1;
end;

procedure TZREvaluator.FindOp3(const strExpr : string; var Op : TZREvOperator; var Pos, Len : integer);
var
  n : integer;
  booFound : boolean;
  intParenteses : integer;
  booString : boolean;
  aString : string[255];
begin
  n := 1;
  intParenteses := 0;
  booFound := false;
  booString := false;
  Len := 1;
  aString := strExpr;
  while (N < Length(strExpr)) and (not booFound) do
  begin
    booFound := true;
    case aString[N] of
      '(' : if not booString then inc(intParenteses);
      ')' : if not booString then dec(intParenteses);
      '''': booString := not booString;
    end;
    if (intParenteses = 0) and (not booString) and (N > 1) then
    begin
      case aString[N] of
        '*' : Op := opMul;
        '/' : Op := opDiv;
        ' ' : if (AnsiLowercase(copy(strExpr, n + 1, 4)) = 'and ') then
              begin
                Op := opAnd;
                Len := 3;
                inc(N);
              end else
                booFound := false;
      else
        booFound := false;
    end;
  end else
    booFound := false;
    inc(N);
  end;
  if booFound then
    Pos := N - 1
  else
    Pos := -1;
end;

function TZREvaluator.NegateResult(const Res : TZREvValue) : TZREvValue;
begin
  result.Kind := Res.Kind;
  case Res.Kind of
    valInteger: Result.intResult := - Res.intResult;
    valDouble : Result.dblResult := -Res.dblResult;
    valBoolean: Result.booResult := not Res.booResult;
  else
    Result.Kind := valError;
  end;
end;

function TZREvaluator.EvalVariable(strVariable : string) : TZREvValue;
var
  SeparatorPos : integer;
  DSName : string;
  FieldName : string;
  aDataSet : TDataSet;
  aField : TField;
  I : integer;
begin
  if assigned(FDataSets) then
  begin
    SeparatorPos := AnsiPos('.', strVariable);
    DSName := AnsiUpperCase(copy(StrVariable, 1, SeparatorPos - 1));
    FieldName := AnsiUpperCase(copy(strVariable, SeparatorPos + 1, length(StrVariable) - SeparatorPos));
    aField := nil;
    aDataSet := nil;
    if length(DSName)>0 then
    begin
      for I := 0 to FDataSets.Count - 1 do
        if AnsiUpperCase(TDataSet(FDataSets[I]).Name) = DSName then
        begin
          aDataSet := TDataSet(FDataSets[I]);
          break;
        end;
      if aDataSet<>nil then
        aField := aDataSet.FindField(FieldName);
    end else
    begin
      for I := 0 to FDataSets.Count - 1 do
        with TDataSet(FDataSets[I]) do
        begin
          aField := FindField(FieldName);
          if aField <> nil then break;
        end;
    end;
    if aField <> nil then
      FiFo.Put(TZREvElementDataField.CreateField(aField))
    else
      FiFo.Put(TZREvElementError.Create);
  end else
    FiFo.Put(TZREvElementError.Create);
end;

function TZREvaluator.EvalString(const strString : string) : TZREvValue;
begin
  result.Kind := valString;
  result.strResult := strString;
  FiFo.Put(TZREvElementString.CreateString(Result.StrResult));
end;

function TZREvaluator.EvalFunction(strFunc : string; const strArg : string) : TZREvValue;
var
  DelimPos : integer;
  aString : string;
  Res : TZREvValue;
  aFunc : TZREvElement;
begin
  StrFunc := AnsiUpperCase(StrFunc);
  aFunc := ZRFunctionLibrary.GetFunction(strFunc);
  if AFunc is TZREvElementError then
  begin
    if StrArg = '' then
    begin
      AFunc.Free;
      EvalVariable(StrFunc)
    end else
      FiFo.Put(AFunc);
  end else
  begin
    FiFo.Put(AFunc);
    if not (aFunc is TZREvElementError) then
    begin
      aString := strArg;
      repeat
        FindDelimiter(aString, DelimPos);
        if DelimPos > 0 then
          Res := Evaluate(copy(aString, 1, DelimPos - 1))
        else
          if length(aString) > 0 then Res := Evaluate(aString);
        Delete(aString, 1, DelimPos);
      until DelimPos = 0;
    end;
    FiFo.Put(TZREvElementArgumentEnd.Create);
  end;
end;

function TZREvaluator.EvalConstant(const strConstant : string) : TZREvValue;
var
  N : integer;
  aString : string[255];
begin
  N := 1;
  aString := strConstant;
  while (N <= Length(aString)) and  (aString[N] in ['0'..'9']) do
    inc(N);
  result.Kind := valInteger;
  while ((N <= Length(aString)) and (aString[N] in ['0'..'9', '.', 'e', 'E', '+', '-'])) do
  begin
    inc(N);
    result.Kind := valDouble;
  end;
  if N - 1 <> Length(aString) then
    result.Kind := valError
  else
  begin
    if result.Kind = valInteger then
      result.intResult := StrToInt(aString)
    else
      if result.Kind = valDouble then
      begin
        if DecimalSeparator <> '.' then
        begin
          while pos('.', aString) > 0 do
            aString[pos('.', aString)] := DecimalSeparator;
        end;
        result.dblResult := StrToFloat(aString);
      end;
  end;
  if result.Kind=valError then
    FiFo.Put(TZREvElementError.Create)
  else
    FiFo.Put(TZREvElementConstant.CreateConstant(Result));
end;

function TZREvaluator.EvalFunctionExpr(const strFunc : string) : TZREvValue;
var
  argRes : TZREvValue;
  po : integer;
begin
  po := AnsiPos('(', StrFunc);
  if po > 0 then
  begin
    if copy(StrFunc, length(StrFunc), 1) = ')' then
    begin
      result := EvalFunction(copy(StrFunc, 1, po - 1), copy(StrFunc, po + 1, length(strFunc) - po - 1));
    end else
    begin
      argRes.Kind := valError;
      result := EvalFunction('', '');
    end;
  end else
  begin
    argRes.Kind := valError;
    result := EvalFunction(StrFunc, '');
  end;
end;

function TZREvaluator.EvalFactor(strFactorExpr : string) : TZREvValue;
var
  aString : string[255];
begin
  TrimString(strFactorExpr);
  aString := strFactorExpr;
  if (AnsiLowerCase(Copy(strFactorExpr, 1, 3)) = 'not') then
    result := NegateResult(EvalFactor(Copy(strFactorExpr, 4, Length(strFactorExpr))))
  else
    case aString[1] of
      '(' : if strFactorExpr[Length(strFactorExpr)] = ')' then
              result := Evaluate(Copy(strFactorExpr, 2, Length(strFactorExpr) - 2))
            else
              result.Kind := valError;
      '-' : result := NegateResult(EvalFactor(Copy(strFactorExpr, 2, Length(strFactorExpr))));
      '+' : result := EvalFactor(Copy(strFactorExpr, 2, Length(strFactorExpr)));
      '0'..'9' : result := EvalConstant(strFactorExpr);
      '''' : if aString[Length(strFactorExpr)] = '''' then
               result := EvalString(Copy(strFactorExpr, 2, Length(strFactorExpr) - 2))
             else
             begin
               result.Kind := valError;
               FiFo.Put(TZREvElementError.Create);
             end;
      '[' : if aString[Length(strFactorExpr)] = ']' then
              result := EvalVariable(Copy(strFactorExpr, 2, Length(strFactorExpr) - 2))
            else
              result.Kind := valError;
      'A'..'Z', 'a'..'z' : result := EvalFunctionExpr(strFactorExpr);
    else
    begin
      FiFo.Put(TZREvElementError.Create);
      result.Kind := valError;
    end;
  end;
end;

function TZREvaluator.EvalSimpleExpr(const strSimplExpr : string) : TZREvValue;
var
  op : TZREvOperator;
  intStart,
  intLen : integer;
  Res1,
  Res2 : TZREvValue;
begin
  FindOp2(strSimplExpr, op, intStart, intLen);
  if intStart > 0 then
  begin
    FiFo.Put(TZREvElementOperator.CreateOperator(Op));
    Res1 := EvalTerm(Copy(strSimplExpr, 1, intStart - 1));
    Res2 := EvalSimpleExpr(Copy(strSimplExpr, intStart + intLen, Length(strSimplExpr)));
  end else
    result := EvalTerm(strSimplExpr);
end;

function TZREvaluator.EvalTerm(const strTermExpr : string) : TZREvValue;
var
  op : TZREvOperator;
  intStart,
  intLen : integer;
  Res1,
  Res2 : TZREvValue;
begin
  FindOp3(strTermExpr, op, intStart, intLen);
  if intStart > 0 then
  begin
    FiFo.Put(TZREvElementOperator.CreateOperator(Op));
    Res1 := EvalFactor(Copy(strTermExpr, 1, intStart - 1));
    Res2 := EvalTerm(Copy(strTermExpr, intStart + intLen, Length(strTermExpr)));
  end else
    result := EvalFactor(strTermExpr);
end;

function TZREvaluator.Evaluate(const strExpr : string) : TZREvValue;
var
  op : TZREvOperator;
  intStart,
  intLen : integer;
  Res1,
  Res2 : TZREvValue;
begin
  FindOp1(strExpr, op, intStart, intLen);
  if intStart > 0 then
  begin
    FiFo.Put(TZREvElementOperator.CreateOperator(Op));
    Res1 := EvalSimpleExpr(Copy(strExpr, 1, intStart - 1));
    Res2 := EvalSimpleExpr(Copy(strExpr, intStart + intLen, Length(strExpr)));
  end else
    result := EvalSimpleExpr(strExpr);
end;

procedure TZREvaluator.Prepare(const strExpr : string);
var
  Value : TZREvValue;
begin
  if Prepared then Unprepare;
  FiFo := TZRFiFo.Create;
  if strExpr = '' then
    Value := Evaluate(''' ''')
  else
    Value := Evaluate(strExpr);
  Prepared := true;
end;

procedure TZREvaluator.UnPrepare;
begin
  if Assigned(FiFo) then
    begin
      FiFo.Free;
      FiFo:= nil;
    end;
  Prepared := false;
end;

procedure TZREvaluator.Reset;
var
  I : integer;
begin
  for I := 0 to FiFo.fFiFo.Count - 1 do
    TZREvElement(FiFo.fFiFo[I]).Reset;
end;

function TZREvaluator.Value : TZREvValue;
begin
  FiFo.Start;
  Result := TZREvElement(FiFo.Get).Value(FiFo);
end;

function TZREvaluator.GetIsAggreg : boolean;
var
  I : integer;
begin
  Result := false;
  for I := 0 to FiFo.fFiFo.Count - 1 do
    Result := Result or TZREvElement(FiFo.fFiFo[I]).IsAggreg;
end;

function TZREvaluator.GetAggregate : boolean;
begin
  Result := FiFo.Aggreg;
end;

procedure TZREvaluator.SetAggregate(Value : boolean);
begin
  FiFo.Aggreg := Value;
end;

function TZREvaluator.Calculate(const strExpr : string) : TZREvValue;
begin
  Prepare(strExpr);
  result := Value;
  UnPrepare;
end;

initialization
  ZRFunctionLibrary := TZRFunctionLibrary.Create;

  { Register ZREvaluator functions }
  RegisterZRFunction(TZREvIfFunction,'IF', 'IF(<Exp>, <X>, <Y>)|' + LoadStr(SzrIfDesc), LoadStr(SzrQSD), '5BVV');
  RegisterZRFunction(TZREvStrFunction, 'STR', 'STR(<X>)|' + LoadStr(SzrStrDesc), LoadStr(SzrQSD), '7N');
  RegisterZRFunction(TZREvUpperFunction, 'UPPER', 'UPPER(<X>)|' + LoadStr(SzrUpperDesc), LoadStr(SzrQSD), '7S');
  RegisterZRFunction(TZREvLowerFunction, 'LOWER', 'LOWER(<X>)|' + LoadStr(SzrLowerDesc), LoadStr(SzrQSD), '7S');
  RegisterZRFunction(TZREvPrettyFunction, 'PRETTY', 'PRETTY(<X>)|' + LoadStr(SzrPrettyDesc), LoadStr(SzrQSD), '7S');
  RegisterZRFunction(TZREvTimeFunction, 'TIME', 'TIME|' + LoadStr(SzrTimeDesc), LoadStr(SzrQSD), '1');
  RegisterZRFunction(TZREvDateFunction, 'DATE', 'DATE|' + LoadStr(SzrDateDesc), LoadStr(SzrQSD), '1');
  RegisterZRFunction(TZREvCopyFunction, 'COPY', 'COPY(<X>, <St>,<Len>)|' + LoadStr(SzrCopyDesc), LoadStr(SzrQSD), '7SNN');
  RegisterZRFunction(TZREvSumFunction, 'SUM', 'SUM(<X>)|' + LoadStr(SzrSumDesc), LoadStr(SzrQSD), '3N');
  RegisterZRFunction(TZREvCountFunction, 'COUNT', 'COUNT|'+ LoadStr(SzrCountDesc), LoadStr(SzrQSD), '3');
  RegisterZRFunction(TZREvMaxFunction, 'MAX', 'MAX(<X>)|' + LoadStr(SzrMaxDesc), LoadStr(SzrQSD), '3V');
  RegisterZRFunction(TZREvMinFunction, 'MIN', 'MIN(<X>)|' + LoadStr(SzrMinDesc), LoadStr(SzrQSD), '3V');
  RegisterZRFunction(TZREvAverageFunction, 'AVERAGE', 'AVERAGE(<X>)|' + LoadStr(SzrAverageDesc), LoadStr(SzrQSD), '3N');
  RegisterZRFunction(TZREvTrue, 'TRUE', 'TRUE|' + LoadStr(SzrBoolDesc), LoadStr(SzrQSD), '5');
  RegisterZRFunction(TZREvFalse, 'FALSE', 'FALSE|' + LoadStr(SzrBoolDesc), LoadStr(SzrQSD), '5');
  RegisterZRFunction(TZREvIntFunction, 'INT', 'INT(<X>)|' + LoadStr(SzrIntDesc), LoadStr(SzrQSD), '2N');
  RegisterZRFunction(TZREvFracFunction, 'FRAC', 'FRAC(<X>)|' + LoadStr(SzrFracDesc), LoadStr(SzrQSD), '2N');
  RegisterZRFunction(TZREvSQRTFunction, 'SQRT', 'SQRT(<X>)|' + LoadStr(SzrSqrtDesc), LoadStr(SzrQSD), '2N');
  RegisterZRFunction(TZREvTypeOfFunction, 'TYPEOF', 'TYPEOF(<Exp>)|' + LoadStr(SzrTypeofDesc), LoadStr(SzrQSD), '6N');
  RegisterZRFunction(TZREvFormatFloatFunction,'FORMATFLOAT', 'FORMATFLOAT(<F>,<N>|' + LoadStr(SzrFormatFloatDesc), LoadStr(SzrQSD), '7SN');
  RegisterZRFunction(TZREvFormatFloatFunction,'FORMATNUMERIC', 'FORMATNUMERIC(<F>,<N>|' + LoadStr(SzrFormatFloatDesc), LoadStr(SzrQSD), '7SN');

finalization
  ZRFunctionLibrary.Free;

end.

